function video = BuildSuperpixelGraphVideo(video, spatial_neigh, temporal_neigh)

    if (nargin < 2 || isempty(spatial_neigh))
        spatial_neigh = 0;
    end
    
    if (nargin < 3 || isempty(temporal_neigh))
        temporal_neigh = 0;
    end
    
    num_frames = size(video.grayscale_frames, 3);
    num_sups = video.sup_idx_range(end,2);
    adjmat_and_affinity = [];
    
    mean_sup_area = video.height * video.width * num_frames / num_sups;
    sup_radius = sqrt(mean_sup_area/pi);
    
    spatial_radius = sup_radius*spatial_neigh;
    temporal_radius = temporal_neigh;
    
    sup_frames = zeros(num_sups, 1);
    for i = 1:num_frames
        sup_frames(video.sup_idx_range(i,1):video.sup_idx_range(i,2)) = i;
    end
    
    disp('Computing spatial superpixel adjacencies based on attachment');

    if (strcmp(video.sup_type, 'lattice'))
        adjmat_lattice = SuperpixelAdjacencyLattice(video.lattice_numH, video.lattice_numV);
        [sup1_offset,sup2_offset] = find(tril(adjmat_lattice, -1));
    end
        
    for i = 1:num_frames
        disp(['Frame ', num2str(i)]);

        if (strcmp(video.sup_type, 'turbopixels'))
            [sup1_offset,sup2_offset] = find(tril(video.adjmat{i}, -1));
        end
        
        sup1 = sup1_offset + video.sup_idx_range(i,1) - 1;
        sup2 = sup2_offset + video.sup_idx_range(i,1) - 1;
                
        % Add spatial affinities
        adjmat_and_affinity = [adjmat_and_affinity; [sup1, sup2, ones(size(sup1))]];
    end

    video.spatial_neighbours = adjmat_and_affinity(:,1:2);
    
    if (temporal_neigh >= 0)
        disp('Computing temporal superpixel adjacencies based on superpixel flow');


        for i = 1:num_frames-1
            disp(['Processing frame ', num2str(i),' out of ', num2str(num_frames)]);
            num_sups_cur = video.sup_idx_range(i,2) - video.sup_idx_range(i,1) + 1;
            for s = 1:num_sups_cur
                s_global = s + video.sup_idx_range(i,1) - 1;

                if (video.supSpeedReliab(s_global))
                    new_center = video.sup_centers(s_global, :) + video.supSpeed(s_global, :);
                else
                    new_center = video.sup_centers(s_global, :);
                end                

                new_center(1) = min(max(round(new_center(1)), 1), video.width); 
                new_center(2) = min(max(round(new_center(2)), 1), video.height); 
                ind = video.superpixels(new_center(2), new_center(1), i+1);

                if (ind == 0)
                    sq_dist_to_next_frame =...
                        (new_center(1) - video.sup_centers(video.sup_idx_range(i+1,1):video.sup_idx_range(i+1,2), 1)).^2 + ...
                        (new_center(2) - video.sup_centers(video.sup_idx_range(i+1,1):video.sup_idx_range(i+1,2), 2)).^2;

                    [min_dist, ind] = min(sq_dist_to_next_frame);
                end

                sup1 = s_global;
                sup2 = video.sup_idx_range(i+1,1) + ind - 1;

                adjmat_and_affinity = [adjmat_and_affinity; [sup1, sup2, ones(size(sup1))]];
            end
        end
    end
    
    if (temporal_neigh == -1)
        disp('Computing temporal superpixel adjacencies based on superpixel flow (multiple connections) ');

        for i = 1:num_frames-1
            disp(['Processing frame ', num2str(i),' out of ', num2str(num_frames)]);
            num_sups_cur = video.sup_idx_range(i,2) - video.sup_idx_range(i,1) + 1;
            Vx = video.Vx(:,:,i);
            Vy = video.Vy(:,:,i);
            reliab = video.flow_reliab(:,:,i);

            for s = 1:num_sups_cur            
                s_global = s + video.sup_idx_range(i,1) - 1;

                pix_idx = (video.superpixels(:,:,i) == s);
                pix_idx = reliab & pix_idx;
                [y,x] = find(pix_idx);
                new_x = x + Vx(pix_idx);
                new_y = y + Vy(pix_idx);
                new_x = min(max(round(new_x), 1), video.width); 
                new_y = min(max(round(new_y), 1), video.height); 

                ind = video.superpixels(sub2ind(size(video.superpixels), ...
                    new_y, new_x, repmat(i+1, size(new_x))));

                if (all(ind == 0))
                    if (video.supSpeedReliab(s_global))
                        new_center = video.sup_centers(s_global, :) + video.supSpeed(s_global, :);
                    else
                        new_center = video.sup_centers(s_global, :);
                    end                

                    new_center(1) = min(max(round(new_center(1)), 1), video.width); 
                    new_center(2) = min(max(round(new_center(2)), 1), video.height); 
                    ind = video.superpixels(new_center(2), new_center(1), i+1);


                    if (ind == 0)
                        sq_dist_to_next_frame =...
                            (new_center(1) - video.sup_centers(video.sup_idx_range(i+1,1):video.sup_idx_range(i+1,2), 1)).^2 + ...
                            (new_center(2) - video.sup_centers(video.sup_idx_range(i+1,1):video.sup_idx_range(i+1,2), 2)).^2;

                        [min_dist, ind] = min(sq_dist_to_next_frame);
                    end

                    weight = 1;
                else
                    unique_sups = unique(ind);
                    unique_sups = unique_sups(~ismember(unique_sups, 0));
                    num_connections = numel(unique_sups);
                    weight = zeros(num_connections, 1);
                    num_good_pix = sum(ind ~= 0);

                    for c = 1:num_connections
                        weight(c) = sum(ind == unique_sups(c)) / num_good_pix;
                    end
                    ind = unique_sups;

                end

                sup1 = repmat(s_global, size(ind));
                sup2 = video.sup_idx_range(i+1,1) + ind - 1;
                sup_affinity = weight;

                adjmat_and_affinity = [adjmat_and_affinity; [sup1, sup2, sup_affinity]];
            end
        end
    end
    
    video.temporal_neighbours = adjmat_and_affinity((size(video.spatial_neighbours,1)+1):end,1:2);
    
    adjmat_sparse = sparse(adjmat_and_affinity(:,1), adjmat_and_affinity(:,2), ...
        adjmat_and_affinity(:,3), num_sups, num_sups);
    adjmat_sparse = adjmat_sparse + adjmat_sparse';

    if (spatial_radius > 0 || temporal_radius > 0)
        disp('Computing superpixel adjacencies based on spatial and temporal radius');
        for s = 1:num_sups
            if (mod(s,1000) == 0)
                disp(['Processed ', num2str(s), ' out of ', num2str(num_sups), ' superpixels']);
            end

            % Compute spatial distances between superpixels
%             cur_sup_center = repmat(video.sup_centers(s,:), [num_sups, 1]);
            if (video.supSpeedReliab(s))
                cur_sup_center = repmat(video.sup_centers(s, :), [num_sups, 1]) + ...
                    repmat((sup_frames(s) - sup_frames), [1, 2]) .* ...
                    repmat(video.supSpeed(s, :), [num_sups, 1]);
            else
                cur_sup_center = repmat(video.sup_centers(s, :), [num_sups, 1]);
            end                
           
            spatial_dist_sq = (sum((cur_sup_center - video.sup_centers).^2, 2));

            % Compute temporal distances between superpixels
            temporal_dist = abs(sup_frames(s) - sup_frames);

            neighbours = (spatial_dist_sq <= spatial_radius^2 & temporal_dist <= max(temporal_radius,0));
            aff = ones(num_sups, 1);            
            neighbours(adjmat_sparse(s, :) > 0) = true;
            aff(adjmat_sparse(s, :) > 0) = adjmat_sparse(s, adjmat_sparse(s, :) > 0);
            adjmat_sparse(s, neighbours) = aff(neighbours);
            %spatial_dist_sq(neighbours);
        end
    end
    
    
    [sup1, sup2, affinity] = find(tril(adjmat_sparse, -1));
    
    video.spatial_neigh = spatial_neigh;
    video.temporal_neigh = temporal_neigh;
    video.adjmat_and_affinity = [sup1, sup2, affinity];