#include <stdio.h>
#include <math.h>
#include "parametric.h"
#include "mex.h"

// print cut C_k (0<=k<=N where N is the number returned by Solve()).
void PrintCut(Parametric* p, int numNodes, int k)
{
	int i;

	printf("Cut (");
	for (i=0; i<numNodes; i++)
	{
		int x = (p->GetRegionCount(p->GetRegion(i)) >= k) ? 1 : 0; // label of node i in C_k
		printf(" %d ", x);
	}
	printf(")");
}

//Assigns the cut C_k to the matrix of returned cuts
void AssignCutToSolution(Parametric* p, int numNodes, int k, mxLogical* selectedNodes)
{
	int i;

	for (i=0; i<numNodes; i++)
	{
		*(selectedNodes + k*numNodes + i) = (p->GetRegionCount(p->GetRegion(i)) >= k); // label of node i in C_k
	}

}
// Minimize energy of the type 
// sum_i((nodeCost(1,i)*lambda+nodeCost(2,i))*X_i) + sum_ij(X_i*X_j*edgeCost(i,j))
//
// Parameters:
//    nodeCost  - unary node costs including lambda's as described above
//    edgeCost  - pairwise costs
//    lambdaMin - minimum lambda for the search
//    lambdaMax - maximum lambda for the search
//
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
		const mxArray *prhs[])
{
	double *nodeCost, *edgeCost, *lambdas;
	mxLogical *selectedNodes;
	int numNodes;
	
	/* The input must be a noncomplex scalar double.*/
	numNodes = (int)mxGetM(prhs[1]);

    printf("Number of nodes: %d \n", numNodes);
    
	/* Assign pointers to each input and output. */
	nodeCost = mxGetPr(prhs[0]);
	edgeCost = mxGetPr(prhs[1]);

    // Compute the number of edges
    int numEdges = 0;
	for (int i = 0; i < numNodes; i++)
	{
		for (int j = 0; j < i; j++)
		{
            
            if (fabs(*(edgeCost + j*numNodes + i)) > 0)
            {
                numEdges++;
            }
        }
    }
    
    printf("Number of edges: %d \n", numEdges);
    
    
    Parametric::RegionId r;
	Parametric* p = new Parametric(/*maximum # of nodes*/ numNodes, /*maximum # of edges*/ numEdges*2); 
	Parametric::REAL lambdaMin = (*mxGetPr(prhs[2]));
	Parametric::REAL lambdaMax = (*mxGetPr(prhs[3]));

    printf("lambdaMin: %f \n", lambdaMin);
    printf("lambdaMax: %f \n", lambdaMax);
    
    p -> AddNode(numNodes); 

	for (int i = 0; i < numNodes; i++)
	{
        p -> AddUnaryTerm(i,  *(nodeCost + i*2 + 0), *(nodeCost + i*2 + 1)); // (nodeCost(1,i)*lambda+nodeCost(2,i))*X_i
        printf("Adding term : (%f * lambda + %f) * X_%d \n", *(nodeCost + i*2 + 0), *(nodeCost + i*2 + 1), i+1);
	}

	for (int i = 0; i < numNodes; i++)
	{
		for (int j = 0; j < i; j++)
		{
            
            if (fabs(*(edgeCost + j*numNodes + i)) > 0)
            {
                if (*(edgeCost + j*numNodes + i) > 0) {                    
                    delete p;
                    mexErrMsgTxt("Irregular terms detected");
                }
                
                p -> AddPairwiseTerm(/* NodeId */ i, /* NodeId */ j,     0, 0, 0, *(edgeCost + j*numNodes + i)*0.5);  // X_i*X_j*edgeCost(i,j)
                p -> AddPairwiseTerm(/* NodeId */ j, /* NodeId */ i,     0, 0, 0, *(edgeCost + j*numNodes + i)*0.5);  // X_i*X_j*edgeCost(i,j)
                printf("Adding term: %f * X_%d * X_%d \n", *(edgeCost + j*numNodes + i), j+1, i+1);
            }
            
		}
	}


    int N = p -> Solve(lambdaMin, lambdaMax);

	plhs[0] = mxCreateLogicalMatrix(numNodes, N+1);
    plhs[1] = mxCreateDoubleMatrix(N, 1, mxREAL);

	selectedNodes = mxGetLogicals(plhs[0]);
    lambdas = mxGetPr(plhs[1]);
	
	r = p->GetFirstRegion();
	if (p->GetRegionLambda(r) > lambdaMin)
	{
		PrintCut(p, numNodes, 0);
		AssignCutToSolution(p, numNodes, 0, selectedNodes);
		printf(" : optimal for lambda's in (-infinity; %f]\n", p->GetRegionLambda(r));
	}

	// Consider cuts C_0,...,C_{N-1}
	for (int k=1; k<N; k++, r=p->GetNextRegion(r))
	{
		lambdas[k-1] = p->GetRegionLambda(r);
		PrintCut(p, numNodes, k);
		AssignCutToSolution(p, numNodes, k, selectedNodes);
		printf(" : optimal for lambda's in [%f; %f]\n", p->GetRegionLambda(r), p->GetRegionLambda(p->GetNextRegion(r)));
	}

	// consider cut C_N
	r = p->GetLastRegion();
	lambdas[N-1] = p->GetRegionLambda(r);
	if (p->GetRegionLambda(r) < lambdaMax)
	{
		PrintCut(p, numNodes, N);
		AssignCutToSolution(p, numNodes, N, selectedNodes);
		printf(" : optimal for lambda's in [%f; +infinity)\n", p->GetRegionLambda(r));
	}
	delete p;

}
